<?php

namespace LavenderBits\Squaby\Mail;

use Kirby\Cms\Page;
use LavenderBits\Squaby\DBService;
use LavenderBits\Squaby\Newsletter;
use LavenderBits\Squaby\Subscriber;

class Mail
{
    public $subject = "";
    public $template = "";
    public $data = "";
    public $recipient;
    private $newsletter;
    public $unsubscribe = null;

    function __construct(
        Newsletter $newsletter,
        Subscriber $recipient,
        string $subject,
        string $template,
        array $data = [],
        string|null $unsubscribe = null
    ) {
        $this->recipient = $recipient;
        $this->subject = $subject;
        $this->template = $template;
        $this->data = $data;
        $this->newsletter = $newsletter;
        $this->unsubscribe = $unsubscribe;
    }

    /**
     * Serializes the Mail object to an array
     * compatible with Kirby's email send feature
     *
     * @return array
     */
    public function serialize(): array
    {
        return [
            "subject" => $this->subject,
            "template" => $this->template,
            "to" => $this->recipient->email()->toString(),
            "toName" => $this->recipient->name()->toString() ?? false,
            // TODO: Switch to transport driven fromName and from
            // We have custom transport settings (squaby.transport instead of email.transport)
            // So let's actually use them
            "from" => $this->newsletter->fromEmail()->toString(),
            "fromName" => $this->newsletter->fromName()->toString(),
            "data" => $this->data,
        ];
    }

    /**
     * Creates a privacy friendly email open tracking link
     * Link may be used inside an <img> in any issue template
     * The link is stored in the sqlite
     *
     * @param Page $issue The issue for which to create a unique link
     *
     * @return string
     */
    private static function createTrackingLink(Page $issue): string
    {
        $db = DBService::getDbConnection();
        $token = uuid();
        $issue = urlencode($issue->id());
        $db->table("tracking")->insert([
            "id" => $token,
            "issue_id" => $issue,
            "opened" => 0,
        ]);

        return join("/", [
            site()->url(),
            "newsletter-api",
            "pixel",
            $issue,
            $token,
        ]);
    }

    /**
     * Creates a new mail for a subscriber with an issue
     *
     * @param Page $issue The issue to be send
     * @param Subscriber $subscriber The receiving subscriber
     *
     * @return Mail
     */
    public static function buildForIssue(
        Page $issue,
        Subscriber $subscriber
    ): Mail {
        /** @var Newsletter */
        $newsletter = $issue->parent();
        $unsubscribe = $subscriber->createUnsubscribeLink();
        $browserLink = $subscriber->createUniqueLink($issue);

        // TODO: Fix title and add subtitle maybe
        return new Mail(
            $issue->parent(),
            $subscriber,
            $issue->title(),
            $issue->template()->name(),
            [
                "page" => $issue,
                "subscriber" => $subscriber,
                "unsubscribe" => $unsubscribe,
                "browserLink" => $browserLink,
                "tracking" =>
                    "<img src=\"" . self::createTrackingLink($issue) . "\" />",
            ],
            $unsubscribe
        );
    }
}
