<?php

namespace LavenderBits\Squaby\Queue;

use Error;
use Kirby\Cms\Page;
use Kirby\Database\Database;
use Kirby\Database\Db;
use Kirby\Toolkit\Date;
use LavenderBits\Squaby\DBService;
use LavenderBits\Squaby\Mail\Mail;
use LavenderBits\Squaby\Mail\Pigeon;
use LavenderBits\Squaby\Newsletter;
use LavenderBits\Squaby\Subscriber;

class Scheduler
{
    /**
     * Schedule a newsletter issue for delivery
     *
     * @param Page $issue The newsletter issue
     * @param Date $deliveryDate The date at which sending should begin
     *
     * @return bool  Run success
     */
    public static function scheduleIssue(Page $issue, Date $deliveryDate): bool
    {
        $db = DBService::getDbConnection();

        // Clear out all previous schedules for this issue
        $db->table("scheduled_issues")->delete(["issue_id", $issue->uuid()]);

        $result = $db->table("scheduled_issues")->insert([
            "issue_id" => $issue->uuid(),
            "deliver_at" => $deliveryDate->getTimestamp(),
        ]);
        return $result !== false;
    }

    /**
     * Queue the sending of all individual e-mails for a given issue that is to be delivered now
     *
     * @param Page $issue The newsletter issue
     *
     * @return bool  Run success
     */
    public static function queueSending(Page $issue): bool
    {
        $db = DBService::getDbConnection();
        /** @var Newsletter */
        $newsletter = $issue->parent();

        // Clear out previous queue entries for this issue
        // TODO: Handle edge case e-mails have already been delivered here
        $db->table("queued_emails")->delete(["issue_id" => $issue->id()]);

        $success = true;
        $subscribers = $newsletter->getActiveSubscribers();
        foreach ($subscribers as $subscriber) {
            /** @var Subscriber $subscriber */
            $result = $db->table("queued_emails")->insert([
                "issue_id" => $issue->uuid(),
                "status" => ScheduleStatus::Scheduled->int(),
                "subscriber_id" => $subscriber->uuid(),
            ]);
            $success = $success && $result !== false;
        }

        return $success;
    }

    /**
     * Queues the sending of all issues that are scheduled for delivery now
     *
     * @return bool  Run success
     */
    public static function queueAll(): bool
    {
        $db = DBService::getDbConnection();
        $scheduledIssues = $db
            ->table("scheduled_issues")
            ->select("*")
            ->where(["status" => ScheduleStatus::Scheduled->int()])
            ->andWhere("deliver_at <= " . Date::now()->getTimestamp())
            ->fetch("array")
            ->all();
        $success = true;
        foreach ($scheduledIssues as $scheduledIssue) {
            $issue = page($scheduledIssue["issue_id"]);
            $result = self::queueSending($issue);
            $success = $success && $result !== false;
            $db->table("scheduled_issues")->update(
                [
                    "status" => ScheduleStatus::Queued->int(),
                ],
                [
                    "id" => $scheduledIssue["id"],
                ]
            );
        }

        return $success;
    }

    /**
     * Updated the status of all queued issues based on delivered emails.
     *
     * @return void
     */
    private static function updateAllQueuedIssues(): void
    {
        $db = DBService::getDbConnection();
        $queuedIssues = $db
            ->table("scheduled_issues")
            ->select("*")
            ->where(["status" => ScheduleStatus::Queued->int()]) // TODO: implement retries
            ->fetch("array")
            ->all();

        foreach ($queuedIssues as $queuedIssue) {
            //Issues will be marked as "delivered" if the amount of queued emails related to them is 0.
            $remainingMails = $db
                ->table("queued_emails")
                ->where(["issue_id" => $queuedIssue["issue_id"]])
                ->andWhere(["status" => ScheduleStatus::Scheduled->int()])
                ->count();

            if ($remainingMails === 0) {
                $db->table("scheduled_issues")->update(
                    ["status" => ScheduleStatus::Delivered->int()],
                    ["id" => $queuedIssue["id"]]
                );
            }
        }
    }

    /**
     * Deliver all scheduled e-mails
     *
     * @return bool  Run success
     */
    public static function deliverAll(): void
    {
        $db = DBService::getDbConnection();
        $queuedEmails = $db
            ->table("queued_emails")
            ->select("*")
            ->where(["status" => ScheduleStatus::Scheduled->int()]) // TODO: implement retries
            ->fetch("array")
            ->all();

        foreach ($queuedEmails as $queuedEmail) {
            $issue = page($queuedEmail["issue_id"]);
            $subscriber = page($queuedEmail["subscriber_id"]);
            $mail = Mail::buildForIssue($issue, $subscriber);
            $db->table("queued_emails")->update(
                ["status" => ScheduleStatus::Delivering->int()],
                ["id" => $queuedEmail["id"]]
            );
            try {
                Pigeon::deliver($mail);
                $db->table("queued_emails")->update(
                    ["status" => ScheduleStatus::Delivered->int()],
                    ["id", $queuedEmail["id"]]
                );
            } catch (Error $e) {
                throw $e;
                $db->table("queued_emails")->update(
                    ["status" => ScheduleStatus::Failed],
                    ["id", $queuedEmail["id"]]
                );
            }
        }
        self::updateAllQueuedIssues();
    }
}
