<?php

namespace LavenderBits\Squaby;

use Kirby\Cms\Page;
use Kirby\Http\Uri;
use Kirby\Toolkit\Date;
use Kirby\Toolkit\Str;
use LavenderBits\Squaby\Mail\Mail;
use LavenderBits\Squaby\Mail\Pigeon;

class Subscriber extends Page
{
    /**
     * Find a subscriber on a newsletter by their secret key
     *
     * @param string $secret
     * @param Newsletter $newsletter
     *
     * @return Subscriber|false The subscriber or false if no subscriber is found
     */
    static function findBySecret(
        string $secret,
        Newsletter $newsletter
    ): Subscriber|false {
        return $newsletter->findSubscriber("secret", $secret);
    }

    /**
     * Marks a subscriber as confirmed & active
     *
     * @return Subscriber  The new state of the subscriber
     */
    public function confirm(): Subscriber
    {
        $newThis = $this->update([
            "hasOptedIn" => true,
            "activeSubscriber" => true,
            "dateConfirmed" => Date::now(),
        ]);
        $newThis = $this->changeStatus("listed");
        return $newThis;
    }

    /**
     * Marks a subscriber as cancelled
     *
     * @return Subscriber  The new state of the subscriber
     */
    public function cancel(): Subscriber
    {
        $newThis = $this;
        $newThis = $newThis->update([
            "activeSubscriber" => false,
            "dateUnsubscribed" => Date::now(),
        ]);
        $newThis = $newThis->changeStatus("unlisted");
        return $newThis;
    }

    /**
     * Delivers the confirmation mail to the subscriber
     *
     * @return bool  true if sending was successful
     */
    public function sendConfirmationMail(): bool
    {
        $mail = new Mail(
            $this->parent(),
            $this,
            $this->parent()->optinMailTitle()->toString(),
            $this->parent()->optinMailTemplate()->toString(),
            ["confirmationLink" => $this->createConfirmationLink()]
        );
        return Pigeon::deliver($mail);
    }

    /**
     * Create a link for confirming a new subscription
     *
     * @return string
     */
    public function createConfirmationLink(): string
    {
        return join("/", [
            site()->url(),
            "newsletter-api/confirm",
            $this->parent()->slug(),
            $this->secret()->toString(),
        ]);
    }

    /**
     * Create a link for cancelling a subscription
     *
     * @return string
     */
    public function createUnsubscribeLink(): string
    {
        return join("/", [
            site()->url(),
            "newsletter-api/unsubscribe",
            $this->parent()->slug(),
            $this->secret()->toString(),
        ]);
    }

    /**
     * Create a link for viewing a newsletter issue in the browser
     *
     * @param Page $issue The issue for which a unique link should be generated
     *
     * @return string
     */
    public function createUniqueLink(Page $issue): string
    {
        return $issue->url() .
            "?sid=" .
            $this->secret()->toString() .
            "&nsl=" .
            urlencode($this->parent()->slug());
    }
}
